/*
 * Driver for USB Creative Labs Nomad Jukebox
 *
 * Tom Vandenhove (tom.vandenhove@ubizen.com)
 *
 * based on nomadII-driver.c by Gabriel Gomiz (gomita@bblanca.com.ar)
 *
 * based on rio500.c by Cesar Miquel (miquel@df.uba.ar)
 *
 * based on hp_scanner.c by David E. Nelson (dnelson@jump.net)
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * Based upon mouse.c (Brad Keryan) and printer.c (Michael Gee).
 *
 */

#include <linux/config.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/signal.h>
#include <linux/sched.h>
#include <linux/errno.h>
#include <linux/miscdevice.h>
#include <linux/random.h>
#include <linux/poll.h>
#include <linux/init.h>
#include <linux/malloc.h>
#include <linux/spinlock.h>
#include <linux/usb.h>
#include <linux/smp_lock.h>

#include "nomadJB.h"

#define NOMADJB_MINOR   96
#define NOMADJB_IF      0

/* Vendor id & Product ids */

/* #define CREATIVE_ID     0x041e */
#define CREATIVE_ID	0x0471
#define CREATIVE_ID2	0x0471

#define NOMADII_ID      0x1002 */
#define NOMADIIMG_ID    0x4004
#define NOMADJB_ID      0x0222
#define NOMADJUKEBOX_ID	0x0222

/* stall/wait timeout for nomad */
#define NAK_TIMEOUT (HZ)

#define NUM_RETRIES 1

/* Size of the nomad input buffer */
#define IBUF_SIZE 0xFE40

/* Size of the nomad output buffer */
#define OBUF_SIZE 0xFE00

/* Endpoint codes */

#define ENDPOINT_CONTROL               0
#define ENDPOINT_INTERRUPT_READ        4
#define ENDPOINT_INTERRUPT_WRITE       3
#define ENDPOINT_BULK_READ             2
#define ENDPOINT_BULK_WRITE            0x82

struct nomadJB_usb_data
{
	struct usb_device *nomadJB_dev;     /* init: probe_nomadJB */
	unsigned int ifnum;                 /* Interface number of the USB device */
	int isopen;                         /* nz if open */
	int present;                        /* Device is present on the bus */
	char *obuf, *ibuf;                  /* transfer buffers */
	char bulk_in_ep, bulk_out_ep;       /* Endpoint assignments */
	wait_queue_head_t wait_q;           /* for timeouts */
	struct semaphore mutex;             /* general race avoidance */
};

static struct nomadJB_usb_data nomadJB_instance;

static int open_nomadJB(struct inode *inode, struct file *file)
{
	struct nomadJB_usb_data *nomadJB = &nomadJB_instance;

	if (nomadJB->isopen || !nomadJB->present) {
		return -EBUSY;
	}

	nomadJB->isopen = 1;

	/* Initialize queue of waiting process */
	init_waitqueue_head(&nomadJB->wait_q);

	MOD_INC_USE_COUNT;

	info("Nomad JB opened.");

	return 0;
}

static int close_nomadJB(struct inode *inode, struct file *file)
{
	struct nomadJB_usb_data *nomadJB = &nomadJB_instance;
	
	nomadJB->isopen = 0;

	MOD_DEC_USE_COUNT;

	info("Nomad JB closed.");

	return 0;
}

static int ioctl_nomadJB(struct inode *inode, struct file *file, unsigned int cmd, unsigned long arg)
{
  	struct NomadJB_cmd nomadJB_cmd;
	struct nomadJB_usb_data *nomadJB = &nomadJB_instance;
	void *data;
	unsigned char *buffer;
	int result, requesttype;
	int retries;
	int errval = 0;

	result = -1;

	/* Sanity check to make sure NomadJB is connected, powered, etc */
	if ( nomadJB == NULL || nomadJB->present == 0 || nomadJB->nomadJB_dev == NULL || nomadJB->isopen == 0 )
		return -EIO;

	switch (cmd) {

	case NOMADJB_RECV_COMMAND:

		data = (void *) arg;

		if (data == NULL)
			break;

		if (copy_from_user(&nomadJB_cmd, data, sizeof(struct NomadJB_cmd)))
			return -EFAULT;

		if (nomadJB_cmd.length > PAGE_SIZE)
			return -EINVAL;

		buffer = (unsigned char *) __get_free_page(GFP_KERNEL);

		if (buffer == NULL)
			return -ENOMEM;

		if (copy_from_user(buffer, nomadJB_cmd.buffer, nomadJB_cmd.length))
			return -EFAULT;

		requesttype = nomadJB_cmd.requesttype | USB_DIR_IN | USB_TYPE_VENDOR | USB_RECIP_OTHER;

		info ("sending command: reqtype=%0x req=%0x value=%0x index=%0x len=%0x", requesttype, 
		      nomadJB_cmd.request, nomadJB_cmd.value, nomadJB_cmd.index, nomadJB_cmd.length);

		/* Send nomadJB control message */
		retries = NUM_RETRIES;

		/* Critical section - START */

		down(&(nomadJB->mutex));

		while (retries) {

			result = usb_control_msg(nomadJB->nomadJB_dev,
						 usb_rcvctrlpipe(nomadJB->nomadJB_dev, ENDPOINT_CONTROL),
						 nomadJB_cmd.request,
						 requesttype,
						 nomadJB_cmd.value,
						 nomadJB_cmd.index, buffer,
						 nomadJB_cmd.length,
						 nomadJB_cmd.timeout);

			if (result == USB_ST_TIMEOUT) {	/* NAK - so hold for a while */

				/* DEBUG */
				info("ioctl_nomadJB: NAK RCV_COMMAND - %d retries left\n",retries);
	
				if (!retries--) {
					/* DEBUG */
					info("ioctl_nomadJB: out of NAK RCV_COMMAND retries - giving up\n");
					errval = -ETIME;
					goto error;
				}

				interruptible_sleep_on_timeout(&nomadJB->wait_q, NAK_TIMEOUT);
				continue;

			}

			else if (result == USB_ST_STALL) {
				
				warn("CLEAR FEATURE requested to remove STALL condition.");

				if (usb_clear_halt(nomadJB->nomadJB_dev,ENDPOINT_CONTROL))
					err("request failed.");
				break;
			}

			else if (result < 0) {
	
				err("Error executing ioctl. code = %d", le32_to_cpu(result));
				retries = 0;
				
			} else {
	
				info("Executed ioctl. Result = %d", le32_to_cpu(result));

				if (copy_to_user(nomadJB_cmd.buffer, buffer, nomadJB_cmd.length)) {
					up(&(nomadJB->mutex));
					free_page((unsigned long) buffer);
					return -EFAULT;
				}
				retries = 0;
				
			}

		}

                /* Critical section - END */
		
		up(&(nomadJB->mutex));

		free_page((unsigned long) buffer);
		break;

	case NOMADJB_SEND_COMMAND:

		data = (void *) arg;

		if (data == NULL)
			break;

		if (copy_from_user(&nomadJB_cmd, data, sizeof(struct NomadJB_cmd)))
			return -EFAULT;

		if (nomadJB_cmd.length > PAGE_SIZE)
			return -EINVAL;

		buffer = (unsigned char *) __get_free_page(GFP_KERNEL);

		if (buffer == NULL)
			return -ENOMEM;

		if (copy_from_user(buffer, nomadJB_cmd.buffer, nomadJB_cmd.length))
			return -EFAULT;

		requesttype = nomadJB_cmd.requesttype | USB_DIR_OUT | USB_TYPE_VENDOR | USB_RECIP_OTHER;

		info("sending command: reqtype=%0x req=%0x value=%0x index=%0x len=%0x",
		     requesttype, nomadJB_cmd.request, nomadJB_cmd.value,
		     nomadJB_cmd.index, nomadJB_cmd.length);

		/* Send nomadJB control message */
		retries = NUM_RETRIES;

                /* Critical section - START */

		down(&(nomadJB->mutex));

		while (retries) {

			result = usb_control_msg(nomadJB->nomadJB_dev,
						 usb_sndctrlpipe(nomadJB->nomadJB_dev, ENDPOINT_CONTROL),
						 nomadJB_cmd.request,
						 requesttype,
						 nomadJB_cmd.value,
						 nomadJB_cmd.index, buffer,
						 nomadJB_cmd.length,
						 nomadJB_cmd.timeout);

			if (result == USB_ST_TIMEOUT) {	/* NAK - so hold for a while */

				/* DEBUG */
				info("ioctl_nomadJB: NAK SEND_COMMAND - %d retries left\n",retries);
	
				if (!retries--) {
					/* DEBUG */
					info("ioctl_nomadJB: out of NAK SEND_COMMAND retries - giving up\n");
					errval = -ETIME;
					goto error;
				}
	
				interruptible_sleep_on_timeout(&nomadJB->wait_q, NAK_TIMEOUT);
				continue;

			}

			else if (result == USB_ST_STALL) {
				
				warn("CLEAR FEATURE requested to remove STALL condition.");
				if (usb_clear_halt(nomadJB->nomadJB_dev,ENDPOINT_CONTROL))
					err("request failed.");
				break;
			}

			else if (result < 0) {
	
				err("Error executing ioctl. code = %d", le32_to_cpu(result));
				retries = 0;

			} else {
	
				info("Executed ioctl. Result = %d", le32_to_cpu(result));
				retries = 0;
	
			}

		}

                /* Critical section - END */
		
		up(&(nomadJB->mutex));

		free_page((unsigned long) buffer);
		break;

	case NOMADJB_RESET_PIPES:
		
                /* Critical section - START */
		down(&(nomadJB->mutex));		
		
		if (usb_clear_halt(nomadJB->nomadJB_dev,usb_rcvbulkpipe(nomadJB->nomadJB_dev,ENDPOINT_BULK_READ))) {
			info("reset pipes request failed.");
			errval = -EPIPE;
			goto error;
			}

		if (usb_clear_halt(nomadJB->nomadJB_dev,usb_sndbulkpipe(nomadJB->nomadJB_dev,ENDPOINT_BULK_WRITE))) {
			info("reset pipes request failed.");
			errval = -EPIPE;
			goto error;
			}

		/* Critical section - END */
		up(&(nomadJB->mutex));

		return 0;

	default:
		return -ENOIOCTLCMD;
		break;
	}

	return result;

error:
	up(&(nomadJB->mutex));
	return errval;

}

static ssize_t write_nomadJB(struct file *file, const char *buffer, size_t count, loff_t * ppos)
{
	struct nomadJB_usb_data *nomadJB = &nomadJB_instance;

	unsigned long copy_size;
	unsigned long bytes_written = 0;
	unsigned int partial;
	unsigned int pipe;

	int errval = 0;
	int result = 0;
	int retries;

	/* Sanity check to make sure nomadJB is connected, powered, etc */
	if ( nomadJB == NULL || nomadJB->present == 0 || nomadJB->nomadJB_dev == NULL || nomadJB->isopen == 0 )
		return -EIO;

	/* Critical section - START */

	down(&(nomadJB->mutex));	

	do {
		unsigned long thistime;
		char *obuf = nomadJB->obuf;
		
		thistime = copy_size = (count >= OBUF_SIZE) ? OBUF_SIZE : count;
		//thistime = copy_size = (count >= 0x10) ? 0x10 : count;
		info("thistime = %ul",thistime);
		if (copy_from_user(nomadJB->obuf, buffer, copy_size)) {
			errval = -EFAULT;
			goto error;
		}
		info("copy_from_user succeeded");
		retries = NUM_RETRIES;

		pipe = usb_sndbulkpipe(nomadJB->nomadJB_dev,ENDPOINT_BULK_WRITE);

		while (thistime) {

			if (!nomadJB->nomadJB_dev) {
				errval = -ENODEV;
				goto error;
			}
			
			if (signal_pending(current)) {
				up(&(nomadJB->mutex));
				return bytes_written ? bytes_written : -EINTR;
			}
			info("starting write");
			result = usb_bulk_msg(nomadJB->nomadJB_dev, pipe,
					      obuf, thistime, &partial, (int) (5*HZ));

			info("write stats: result:%d thistime:%lu partial:%u", result, thistime, partial);

			if (result == USB_ST_TIMEOUT) {	/* NAK - so hold for a while */

				/* DEBUG */
				info("write_nomadJB: NAK  - %d retries left\n",retries);
	
				if (!retries--) {
					/* DEBUG */
					info("write_nomadJB: out of NAK retries - giving up\n");
					errval = -ETIME;
					goto error;
				}

				/* DEBUG */
				warn("CLEAR FEATURE requested to remove STALL condition.");
				if (usb_clear_halt(nomadJB->nomadJB_dev,pipe))
					err("request failed.");
	
				interruptible_sleep_on_timeout(&nomadJB->wait_q, NAK_TIMEOUT);
				continue;

			} else if (result == USB_ST_STALL) {
				
				warn("CLEAR FEATURE requested to remove STALL condition.");
				if (usb_clear_halt(nomadJB->nomadJB_dev,pipe))
				   err("request failed.");
				break;
				
			} else if (!result && partial) {
	
				obuf += partial;
				thistime -= partial;

			} else
	
				break;
		};
	
		if (result) {

			err("write_nomadJB: Whoops - %x", result);
			errval = -EIO;
			goto error;
	
		}

		bytes_written += copy_size;
		count -= copy_size;
		buffer += copy_size;
		
	} while (count > 0);

	/* Critical section - END */
		
	up(&(nomadJB->mutex));
		
	return bytes_written ? bytes_written : -EIO;

 error:
	up(&(nomadJB->mutex));
	return errval;

}


static ssize_t read_nomadJB(struct file *file, char *buffer, size_t count, loff_t * ppos)
{
	struct nomadJB_usb_data *nomadJB = &nomadJB_instance;

	ssize_t read_count;
	unsigned int partial;
	unsigned int pipe;
	int this_read;
	int result;
	int retries;
	int errval = 0;
	char *ibuf = nomadJB->ibuf;

	/* Sanity check to make sure Nomad JB is connected, powered, etc */
	if ( nomadJB == NULL || nomadJB->present == 0 || nomadJB->nomadJB_dev == NULL || nomadJB->isopen==0 )
		return -EIO;

	/* Do the reading... */
	read_count = 0;

	retries = NUM_RETRIES;

        /* Critical section - START */

	down(&(nomadJB->mutex));

	pipe = usb_rcvbulkpipe(nomadJB->nomadJB_dev,ENDPOINT_BULK_READ);

	while (count > 0) {

		if (signal_pending(current)) {
			errval = read_count ? read_count : -EINTR;
			goto error;
		}

		/* Sanity check */
		if ( !nomadJB->nomadJB_dev ) {
			errval = -ENODEV;
			goto error;
		}

		/* How much are we reading this time? */
		this_read = (count >= IBUF_SIZE) ? IBUF_SIZE : count;

		result = usb_bulk_msg(nomadJB->nomadJB_dev,pipe,
				      ibuf, this_read, &partial,(int) (5*HZ));

		info("read stats: result:%d this_read:%u partial:%u",
		     result, this_read, partial);

		if (partial) {

			count = this_read = partial;

		} else if (result == USB_ST_TIMEOUT) {

			/* DEBUG */
			info("read_nomadJB: NAK - %d retries left\n",retries);

			if (!retries--) {	
                                /* DEBUG */
				info("read_nomadJB: out of NAK retries - giving up\n");
				errval = -ETIME;
				goto error;
			}

			/* DEBUG */
			warn("CLEAR FEATURE requested to remove STALL condition.");
			if (usb_clear_halt(nomadJB->nomadJB_dev,pipe))
				err("request failed."); 

			/* Go wait some time..... and then retry it... */
			interruptible_sleep_on_timeout(&nomadJB->wait_q, NAK_TIMEOUT);
			continue;

		} else if (result == USB_ST_STALL) {
				
			warn("CLEAR FEATURE requested to remove STALL condition.");
			if (usb_clear_halt(nomadJB->nomadJB_dev,pipe))
			   err("request failed.");
			break;

		} else if (result != USB_ST_DATAUNDERRUN) {
			
			err("read_nomadJB: whoops - result:%u partial:%u this_read:%u",
			    result, partial, this_read);
			errval = -EIO;
			goto error;
			
		} else {
			up(&(nomadJB->mutex));
			return (0);

		} /* end if (partial) */
		
		if (this_read) {

			if (copy_to_user(buffer, ibuf, this_read)) {
				errval = -EFAULT;
				goto error;
			}

			count -= this_read;
			read_count += this_read;
			buffer += this_read;

		}

	}
	up(&(nomadJB->mutex));
	return read_count;

error:
	up(&(nomadJB->mutex));
	return errval;

}


static void *probe_nomadJB(struct usb_device *dev, unsigned int ifnum,
			   const struct usb_device_id *id)
{
	struct nomadJB_usb_data *nomadJB = &nomadJB_instance;

	info("USB Creative Nomad JB support registered.");

	if (dev->descriptor.idVendor != CREATIVE_ID2 ) {
		info("not creative ID...");
	} else {
		info("creative ID matches...");
		if (dev->descriptor.idProduct != NOMADJUKEBOX_ID ) {
			info ("not jukebox...");
		} else {
			info ("it's the jukebox...");
			}
		
	}


	if (dev->descriptor.idVendor != CREATIVE_ID ) {

		return NULL;

	}

	info("it's creative!");

	switch (dev->descriptor.idProduct) {

	case NOMADJB_ID:

		info("Creative Labs Nomad JB found at address %d", dev->devnum);
		break;

	case NOMADIIMG_ID:

		info("Creative Labs Nomad II MG found at address %d", dev->devnum);
		break;

	default:
		warn(KERN_INFO "Nomad player model not supported/tested.");
		return NULL;
	 
	}

	/* We don't handle multiple configurations */
	if (dev->descriptor.bNumConfigurations != 1) {
		
		warn(KERN_INFO "Nomad JB multiple configurations is not supported/tested.");
		return NULL;
	}

	if (ifnum != NOMADJB_IF) {
		
		warn(KERN_INFO "Nomad JB interface not supported/tested.");
		return NULL;

	}

	nomadJB->present = 1;
	nomadJB->nomadJB_dev = dev;

	/* Try to set configuration */
	if (usb_set_configuration (dev, dev->config[0].bConfigurationValue) < 0) {

		err("set_configuration failed");
		goto reject;

	}

	info("bound to configuration: %d - num_interfaces %d",dev->config[0].bConfigurationValue,
	     dev->config[0].bNumInterfaces);

	/* Try to set interface */
	if (usb_set_interface (dev, NOMADJB_IF, 0) < 0) {

		err("set_interface failed");
		goto reject;

	}

	info("bound to interface: %d", NOMADJB_IF);

	if (!(nomadJB->obuf = (char *) kmalloc(OBUF_SIZE, GFP_KERNEL))) {

		err("probe_nomadJB: Not enough memory for the output buffer");
		goto reject;

	}

	dbg("probe_nomadJB: obuf address:%p", nomadJB->obuf);

	if (!(nomadJB->ibuf = (char *) kmalloc(IBUF_SIZE, GFP_KERNEL))) {

		err("probe_nomadJB: Not enough memory for the input buffer");
		kfree(nomadJB->obuf);
		goto reject;

	}

	dbg("probe_nomadJB: ibuf address:%p", nomadJB->ibuf);

	init_MUTEX(&(nomadJB->mutex));

	return nomadJB;

 reject:
	nomadJB->nomadJB_dev = NULL;
	return NULL;
}

static void disconnect_nomadJB(struct usb_device *dev, void *ptr)
{
	struct nomadJB_usb_data *nomadJB = (struct nomadJB_usb_data *) ptr;

	if (nomadJB->isopen) {

		nomadJB->isopen = 0;

		/* better let it finish - the release will do what's needed */
		nomadJB->nomadJB_dev = NULL;
		return;

	}

	kfree(nomadJB->ibuf);
	kfree(nomadJB->obuf);

	info("USB Nomad JB disconnected.");

	nomadJB->present = 0;
}

static struct file_operations usb_nomadJB_fops =
{
	read:		read_nomadJB,
	write:	        write_nomadJB,
	ioctl:	        ioctl_nomadJB,
	open:		open_nomadJB,
	release:	close_nomadJB,
};

static struct usb_driver nomadJB_driver =
{
	name:		"nomadJB",
	probe:		probe_nomadJB,
	disconnect:	disconnect_nomadJB,
	fops:		&usb_nomadJB_fops,
	minor:		NOMADJB_MINOR
};

static int __init usb_nomadJB_init(void)
{
	if (usb_register(&nomadJB_driver) < 0)
		return -1;
	
	info("USB Creative Nomad JB support registered.");

	return 0;
}


static void __exit usb_nomadJB_cleanup(void)
{

	struct nomadJB_usb_data *nomadJB = &nomadJB_instance;

	nomadJB->present = 0;

	usb_deregister(&nomadJB_driver);

}

/* --------------------------------------------------------------------- */

MODULE_AUTHOR ("Tom Vandenhove, tom.vandenhove@ubizen.com");
MODULE_DESCRIPTION ("USB Creative Nomad JB Driver for Linux (c)2000");

module_init(usb_nomadJB_init);
module_exit(usb_nomadJB_cleanup);

/* --------------------------------------------------------------------- */
